from .audioAspectsRegistry import registrationAudioAspect, cacheAudioAnalyzers, analyzeAudioFile, \
	analyzeAudioListPathFilenames, getListAvailableAudioAspects, audioAspects

__all__ = [
	'analyzeAudioFile',
	'analyzeAudioListPathFilenames',
	'audioAspects',
	'getListAvailableAudioAspects',
]

from . import analyzersUseFilename
from . import analyzersUseSpectrogram
from . import analyzersUseTensor
from . import analyzersUseWaveform



from analyzeAudio import registrationAudioAspect, audioAspects, cacheAudioAnalyzers
from typing import Any
import librosa
import numpy
from optype.numpy import ToArray2D, AnyFloatingDType
import cachetools

@cachetools.cached(cache=cacheAudioAnalyzers)
@registrationAudioAspect('Tempogram')
def analyzeTempogram(waveform: ToArray2D[AnyFloatingDType], sampleRate: int, **keywordArguments: Any) -> numpy.ndarray:
	return librosa.feature.tempogram(y=waveform, sr=sampleRate, **keywordArguments)

# "RMS value from audio samples is faster ... However, ... spectrogram ... more accurate ... because ... windowed"
@registrationAudioAspect('RMS from waveform')
def analyzeRMS(waveform: ToArray2D[AnyFloatingDType], **keywordArguments: Any) -> numpy.ndarray:
	arrayRMS = librosa.feature.rms(y=waveform, **keywordArguments)
	return 20 * numpy.log10(arrayRMS, where=(arrayRMS != 0)) # dB

@registrationAudioAspect('Tempo')
def analyzeTempo(waveform: ToArray2D[AnyFloatingDType], sampleRate: int, **keywordArguments: Any) -> numpy.ndarray:
	tempogram = audioAspects['Tempogram']['analyzer'](waveform, sampleRate)
	return librosa.feature.tempo(y=waveform, sr=sampleRate, tg=tempogram, **keywordArguments)

@registrationAudioAspect('Zero-crossing rate') # This is distinct from 'Zero-crossings rate'
def analyzeZeroCrossingRate(waveform: ToArray2D[AnyFloatingDType], **keywordArguments: Any) -> numpy.ndarray:
	return librosa.feature.zero_crossing_rate(y=waveform, **keywordArguments)



import pathlib
import pytest

pathDataSamples = pathlib.Path(__file__).parent / "dataSamples"

listOfFiles = ["pink-20RMS60sec.wav",
"pink-40RMS60sec.wav",
"pink-60RMS60sec.wav",
"testParkMono96kHz32float12.1sec.wav",
"testPink2ch7.1sec.wav",
"testSine2ch5sec.wav",
"testSine2ch5secCopy1.wav",
"testTrain2ch48kHz6.3sec.wav",
"testVideo11sec.mkv",
"testWooWooMono16kHz32integerClipping9sec.wav",
]



import pytest
from Z0Z_tools.pytestForYourUse import PytestFor_defineConcurrencyLimit, PytestFor_oopsieKwargsie

@pytest.mark.parametrize("nameOfTest,callablePytest", PytestFor_defineConcurrencyLimit())
def testConcurrencyLimit(nameOfTest, callablePytest):
	callablePytest()

@pytest.mark.parametrize("nameOfTest,callablePytest", PytestFor_oopsieKwargsie())
def testOopsieKwargsie(nameOfTest, callablePytest):
	callablePytest()



from collections import defaultdict
from typing import Any, cast, NamedTuple
import json
import numpy

# NOTE hey! hey! hey!
# Is blackdetect broken?
# 1. You don't have pytest tests for anything in the entire fricken package
# 2. You tried to improve the blackdetect code, but you didn't test it with anything
# 3. Search for "uncommentToFixBlackdetect"
# NOTE You changed the code because a static type checker was mad at you. Ask yourself,
# "Are you the tool or is the type checker the tool?"

class Blackdetect(NamedTuple):
	black_start: float | None = None
	black_end: float | None = None

def pythonizeFFprobe(FFprobeJSON_utf8: str):
	FFroot: dict[str, Any] = json.loads(FFprobeJSON_utf8)
	Z0Z_dictionaries: dict[str, numpy.ndarray[Any, Any] | dict[str, numpy.ndarray[Any, Any]]] = {}
	if 'packets_and_frames' in FFroot: # Divide into 'packets' and 'frames'
		FFroot = defaultdict(list, FFroot)
		for packetOrFrame in FFroot['packets_and_frames']:
			if 'type' in packetOrFrame:
				FFroot[section := packetOrFrame['type'] + 's'].append(packetOrFrame)
				del FFroot[section][-1]['type']
			else:
				raise ValueError("'packets_and_frames' for the win!")
		del FFroot['packets_and_frames']

	Z0Z_register = [
		'aspectralstats',
		'astats',
		'r128',
		'signalstats',
	]
	leftCrumbs = False
	if 'frames' in FFroot:
		leftCrumbs = False
		# listTuplesBlackdetect = [] # uncommentToFixBlackdetect
		listTuplesBlackdetect: list[Blackdetect] = []
		for indexFrame, FFframe in enumerate(FFroot['frames']):
			if 'tags' in FFframe:
				if 'lavfi.black_start' in FFframe['tags']:
					# listTuplesBlackdetect.append(float(FFframe['tags']['lavfi.black_start'])) # uncommentToFixBlackdetect
					listTuplesBlackdetect.append(Blackdetect(black_start=float(FFframe['tags']['lavfi.black_start'])))
					del FFframe['tags']['lavfi.black_start']
				if 'lavfi.black_end' in FFframe['tags']:
					# listTuplesBlackdetect[-1] = (listTuplesBlackdetect[-1], float(FFframe['tags']['lavfi.black_end'])) # uncommentToFixBlackdetect
					tupleBlackdetectLast = listTuplesBlackdetect.pop() if listTuplesBlackdetect else Blackdetect()
					match tupleBlackdetectLast.black_end:
						case None:
							listTuplesBlackdetect.append(Blackdetect(tupleBlackdetectLast.black_start, float(FFframe['tags']['lavfi.black_end'])))
						case _:
							if tupleBlackdetectLast.black_start is not None:
								listTuplesBlackdetect.append(tupleBlackdetectLast)
							listTuplesBlackdetect.append(Blackdetect(black_end=(float(FFframe['tags']['lavfi.black_end']))))
					del FFframe['tags']['lavfi.black_end']

				# This is not the way to do it
				for keyName, keyValue in FFframe['tags'].items():
					if 'lavfi' in (keyNameDeconstructed := keyName.split('.'))[0]:
						channel = None
						if (registrant := keyNameDeconstructed[1]) in Z0Z_register:
							keyNameDeconstructed = keyNameDeconstructed[2:]
							if keyNameDeconstructed[0].isdigit():
								channel = int(keyNameDeconstructed[0])
								keyNameDeconstructed = keyNameDeconstructed[1:]
							statistic = '.'.join(keyNameDeconstructed)
							if channel is None:
								while True:
									try:
										Z0Z_dictionaries[registrant][statistic][indexFrame] = float(keyValue)
										break  # If successful, exit the loop
									except KeyError:
										if registrant not in Z0Z_dictionaries:
											Z0Z_dictionaries[registrant] = {}
										elif statistic not in Z0Z_dictionaries[registrant]:
											Z0Z_dictionaries[registrant][statistic] = numpy.zeros(len(FFroot['frames']))
										else:
											raise  # Re-raise the exception
							else:
								while True:
									try:
										Z0Z_dictionaries[registrant][statistic][channel - 1, indexFrame] = float(keyValue)
										break  # If successful, exit the loop
									except KeyError:
										if registrant not in Z0Z_dictionaries:
											Z0Z_dictionaries[registrant] = {}
										elif statistic not in Z0Z_dictionaries[registrant]:
												# NOTE (as of this writing) `registrar` can only understand the generic class `numpy.ndarray` and not more specific typing
												valueSherpa = cast(numpy.ndarray, numpy.zeros((channel, len(FFroot['frames']))))
												Z0Z_dictionaries[registrant][statistic] = valueSherpa
										else:
											raise  # Re-raise the exception
									except IndexError:
										if channel > Z0Z_dictionaries[registrant][statistic].shape[0]:
											Z0Z_dictionaries[registrant][statistic] = numpy.resize(Z0Z_dictionaries[registrant][statistic], (channel, len(FFroot['frames'])))
											# Z0Z_dictionaries[registrant][statistic].resize((channel, len(FFroot['frames'])))
										else:
											raise  # Re-raise the exception

				if not FFframe['tags']: # empty = False
					del FFframe['tags']
			if FFframe:
				leftCrumbs = True
		if listTuplesBlackdetect:
			# 2025-03-06: I am _shocked_ that I was able to create a numpy structured array whenever it was that I originally wrote this code.
			arrayBlackdetect = numpy.array(
				[(
					-1.0 if detect.black_start is None else detect.black_start,
					-1.0 if detect.black_end is None else detect.black_end
				) for detect in listTuplesBlackdetect],
				dtype=[('black_start', numpy.float64), ('black_end', numpy.float64)],
				copy=False
			)
			Z0Z_dictionaries['blackdetect'] = arrayBlackdetect
			# Z0Z_dictionaries['blackdetect'] = numpy.array(listTuplesBlackdetect, dtype=[('black_start', numpy.float32), ('black_end', numpy.float32)], copy=False) # uncommentToFixBlackdetect
	if not leftCrumbs:
		del FFroot['frames']
	return FFroot, Z0Z_dictionaries



from collections.abc import Callable, Sequence
from concurrent.futures import ProcessPoolExecutor, as_completed
from numpy.typing import NDArray
from typing import Any, cast, ParamSpec, TypeAlias, TYPE_CHECKING, TypeVar
from Z0Z_tools import defineConcurrencyLimit, oopsieKwargsie, stft
import cachetools
import inspect
import librosa
import multiprocessing
import numpy
from os import PathLike
import pathlib
import soundfile
import torch
import warnings

if TYPE_CHECKING:
	from typing import TypedDict
else:
	TypedDict = dict

if __name__ == '__main__':
	multiprocessing.set_start_method('spawn')

warnings.filterwarnings('ignore', category=UserWarning, module='torchmetrics', message='.*fast=True.*')

parameterSpecifications = ParamSpec('parameterSpecifications')
typeReturned = TypeVar('typeReturned')

audioAspect: TypeAlias = str
class analyzersAudioAspects(TypedDict):
	analyzer: Callable[..., Any]
	analyzerParameters: list[str]

audioAspects: dict[audioAspect, analyzersAudioAspects] = {}
"""A register of 1) measurable aspects of audio data, 2) analyzer functions to measure audio aspects, 3) and parameters of analyzer functions."""

def registrationAudioAspect(aspectName: str) -> Callable[[Callable[parameterSpecifications, typeReturned]], Callable[parameterSpecifications, typeReturned]]:
	"""
	A function to "decorate" a registrant-analyzer function and the aspect of audio data it can analyze.

	Parameters:
		aspectName: The audio aspect that the registrar will enter into the register, `audioAspects`.
	"""

	def registrar(registrant: Callable[parameterSpecifications, typeReturned]) -> Callable[parameterSpecifications, typeReturned]:
		"""
		`registrar` updates the registry, `audioAspects`, with 1) the analyzer function, `registrant`, 2) the analyzer function's parameters, and 3) the aspect of audio data that the analyzer function measures.

		Parameters:
			registrant: The function that analyzes an aspect of audio data.

		Note:
			`registrar` does not change the behavior of `registrant`, the analyzer function.
		"""
		audioAspects[aspectName] = {
			'analyzer': registrant,
			'analyzerParameters': inspect.getfullargspec(registrant).args
		}

		# if registrant.__annotations__.get('return') is not None and issubclass(registrant.__annotations__['return'], subclassTarget): # maybe someday I will understand why this doesn't work
		# if registrant.__annotations__.get('return') is not None and issubclass(registrant.__annotations__.get('return', type(None)), subclassTarget): # maybe someday I will understand why this doesn't work
		if isinstance(registrant.__annotations__.get('return', type(None)), type) and issubclass(registrant.__annotations__.get('return', type(None)), numpy.ndarray): # maybe someday I will understand what all of this statement means
			def registrationAudioAspectMean(*arguments: parameterSpecifications.args, **keywordArguments: parameterSpecifications.kwargs) -> numpy.floating[Any]:
				"""
				`registrar` updates the registry with a new analyzer function that calculates the mean of the analyzer's numpy.ndarray result.

				Returns:
					mean: Mean value of the analyzer's numpy.ndarray result.
				"""
				aspectValue = registrant(*arguments, **keywordArguments)
				return numpy.mean(cast(NDArray[Any], aspectValue))
				# return aspectValue.mean()
			audioAspects[f"{aspectName} mean"] = {
				'analyzer': registrationAudioAspectMean,
				'analyzerParameters': inspect.getfullargspec(registrant).args
			}
		return registrant
	return registrar

def analyzeAudioFile(pathFilename: str | PathLike[Any], listAspectNames: list[str]) -> list[str | float | NDArray[Any]]:
	"""
	Analyzes an audio file for specified aspects and returns the results.

	Parameters:
		pathFilename: The path to the audio file to be analyzed.
		listAspectNames: A list of aspect names to analyze in the audio file.

	Returns:
		listAspectValues: A list of analyzed values in the same order as `listAspectNames`.
	"""
	pathlib.Path(pathFilename).stat() # raises FileNotFoundError if the file does not exist
	dictionaryAspectsAnalyzed: dict[str, str | float | NDArray[Any]] = {aspectName: 'not found' for aspectName in listAspectNames}
	"""Despite returning a list, use a dictionary to preserve the order of the listAspectNames.
	Similarly, 'not found' ensures the returned list length == len(listAspectNames)"""

	with soundfile.SoundFile(pathFilename) as readSoundFile:
		sampleRate: int = readSoundFile.samplerate
		waveform = readSoundFile.read(dtype='float32').astype(numpy.float32)
		waveform = waveform.T

	# I need "lazy" loading
	tryAgain = True
	while tryAgain: # `tenacity`?
		try:
			tensorAudio = torch.from_numpy(waveform)  # memory-sharing
			tryAgain = False
		except RuntimeError as ERRORmessage:
			if 'negative stride' in str(ERRORmessage):
				waveform = waveform.copy()  # not memory-sharing
				tryAgain = True
			else:
				raise ERRORmessage

	spectrogram = stft(waveform, sampleRate=sampleRate)
	spectrogramMagnitude = numpy.absolute(spectrogram)
	spectrogramPower = spectrogramMagnitude ** 2

	pytorchOnCPU = not torch.cuda.is_available()  # False if GPU available, True if not

	for aspectName in listAspectNames:
		if aspectName in audioAspects:
			analyzer = audioAspects[aspectName]['analyzer']
			analyzerParameters = audioAspects[aspectName]['analyzerParameters']
			dictionaryAspectsAnalyzed[aspectName] = analyzer(*map(vars().get, analyzerParameters))

	return [dictionaryAspectsAnalyzed[aspectName] for aspectName in listAspectNames]

def analyzeAudioListPathFilenames(listPathFilenames: Sequence[str] | Sequence[PathLike[Any]], listAspectNames: list[str], CPUlimit: int | float | bool | None = None) -> list[list[str | float | NDArray[Any]]]:
	"""
	Analyzes a list of audio files for specified aspects of the individual files and returns the results.

	Parameters:
		listPathFilenames: A list of paths to the audio files to be analyzed.
		listAspectNames: A list of aspect names to analyze in each audio file.
		CPUlimit (gluttonous resource usage): whether and how to limit the CPU usage. See notes for details.

	Returns:
		rowsListFilenameAspectValues: A list of lists, where each inner list contains the filename and
		analyzed values corresponding to the specified aspects, which are in the same order as `listAspectNames`.

	You can save the data with `Z0Z_tools.dataTabularTOpathFilenameDelimited()`.
	For example,

	```python
	dataTabularTOpathFilenameDelimited(
		pathFilename = pathFilename,
		tableRows = rowsListFilenameAspectValues, # The return of this function
		tableColumns = ['File'] + listAspectNames # A parameter of this function
	)
	```

	Nevertheless, I aspire to improve `analyzeAudioListPathFilenames` by radically improving the structure of the returned data.

	Limits on CPU usage CPUlimit:
		False, None, or 0: No limits on CPU usage; uses all available CPUs. All other values will potentially limit CPU usage.
		True: Yes, limit the CPU usage; limits to 1 CPU.
		Integer >= 1: Limits usage to the specified number of CPUs.
		Decimal value (float) between 0 and 1: Fraction of total CPUs to use.
		Decimal value (float) between -1 and 0: Fraction of CPUs to *not* use.
		Integer <= -1: Subtract the absolute value from total CPUs.

	"""
	rowsListFilenameAspectValues: list[list[str | float | NDArray[Any]]] = []

	if not (CPUlimit is None or isinstance(CPUlimit, (bool, int, float))):
		CPUlimit = oopsieKwargsie(CPUlimit)
	max_workers = defineConcurrencyLimit(CPUlimit)

	with ProcessPoolExecutor(max_workers=max_workers) as concurrencyManager:
		dictionaryConcurrency = {concurrencyManager.submit(analyzeAudioFile, pathFilename, listAspectNames)
									: pathFilename
									for pathFilename in listPathFilenames}

		for claimTicket in as_completed(dictionaryConcurrency):
			cacheAudioAnalyzers.pop(dictionaryConcurrency[claimTicket], None)
			listAspectValues = claimTicket.result()
			rowsListFilenameAspectValues.append(
				[str(pathlib.PurePath(dictionaryConcurrency[claimTicket]).as_posix())]
				+ listAspectValues)

	return rowsListFilenameAspectValues

def getListAvailableAudioAspects() -> list[str]:
	"""
	Returns a sorted list of audio aspect names. All valid values for the parameter `listAspectNames`, for example,
	are returned by this function.

	Returns:
		listAvailableAudioAspects: The list of aspect names registered in `audioAspects`.
	"""
	return sorted(audioAspects.keys())

cacheAudioAnalyzers: cachetools.LRUCache[str | PathLike[Any], NDArray[Any]] = cachetools.LRUCache(maxsize=256)



from analyzeAudio import registrationAudioAspect
from torchmetrics.functional.audio.srmr import speech_reverberation_modulation_energy_ratio
from typing import Any
import numpy
import torch

@registrationAudioAspect('SRMR')
def analyzeSRMR(tensorAudio: torch.Tensor, sampleRate: int, pytorchOnCPU: bool | None, **keywordArguments: Any) -> numpy.ndarray:
	keywordArguments['fast'] = keywordArguments.get('fast') or pytorchOnCPU or None
	return torch.Tensor.numpy(speech_reverberation_modulation_energy_ratio(tensorAudio, sampleRate, **keywordArguments))



from .pythonator import pythonizeFFprobe
from analyzeAudio import registrationAudioAspect, cacheAudioAnalyzers
from os import PathLike
from statistics import mean
from typing import Any, cast
import cachetools
import numpy
import pathlib
import re as regex
import subprocess

@registrationAudioAspect('SI-SDR mean')
def getSI_SDRmean(pathFilenameAlpha: str | PathLike[Any], pathFilenameBeta: str | PathLike[Any]) -> float | None:
	"""
	Calculate the mean Scale-Invariant Signal-to-Distortion Ratio (SI-SDR) between two audio files.
	This function uses FFmpeg to compute the SI-SDR between two audio files specified by their paths.
	The SI-SDR values are extracted from the FFmpeg output and their mean is calculated.
	Parameters:
		pathFilenameAlpha: Path to the first audio file.
		pathFilenameBeta: Path to the second audio file.
	Returns:
		SI_SDRmean: The mean SI-SDR value in decibels (dB).
	Raises:
		subprocess.CalledProcessError: If the FFmpeg command fails.
		ValueError: If no SI-SDR values are found in the FFmpeg output.
	"""
	commandLineFFmpeg = [
		'ffmpeg', '-hide_banner', '-loglevel', '32',
		'-i', f'{str(pathlib.Path(pathFilenameAlpha))}', '-i', f'{str(pathlib.Path(pathFilenameBeta))}',
		'-filter_complex', '[0][1]asisdr', '-f', 'null', '-'
	]
	systemProcessFFmpeg = subprocess.run(commandLineFFmpeg, check=True, stderr=subprocess.PIPE)

	stderrFFmpeg = systemProcessFFmpeg.stderr.decode()

	regexSI_SDR = regex.compile(r"^\[Parsed_asisdr_.* (.*) dB", regex.MULTILINE)

	listMatchesSI_SDR = regexSI_SDR.findall(stderrFFmpeg)
	SI_SDRmean = mean(float(match) for match in listMatchesSI_SDR)
	return SI_SDRmean

@cachetools.cached(cache=cacheAudioAnalyzers)
def ffprobeShotgunAndCache(pathFilename: str | PathLike[Any]) -> dict[str, float | numpy.ndarray]:
	# for lavfi amovie/movie, the colons after driveLetter letters need to be escaped twice.
	pFn = pathlib.PureWindowsPath(pathFilename)
	lavfiPathFilename = pFn.drive.replace(":", "\\\\:")+pathlib.PureWindowsPath(pFn.root,pFn.relative_to(pFn.anchor)).as_posix()

	filterChain: list[str] = []
	filterChain += ["astats=metadata=1:measure_perchannel=Crest_factor+Zero_crossings_rate+Dynamic_range:measure_overall=all"]
	filterChain += ["aspectralstats"]
	filterChain += ["ebur128=metadata=1:framelog=quiet"]

	entriesFFprobe = ["frame_tags"]

	commandLineFFprobe = [
		"ffprobe", "-hide_banner",
		"-f", "lavfi", f"amovie={lavfiPathFilename},{','.join(filterChain)}",
		"-show_entries", ':'.join(entriesFFprobe),
		"-output_format", "json=compact=1",
	]

	systemProcessFFprobe = subprocess.Popen(commandLineFFprobe, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
	stdoutFFprobe, _DISCARDstderr = systemProcessFFprobe.communicate()
	FFprobeStructured = pythonizeFFprobe(stdoutFFprobe.decode('utf-8'))[-1]

	dictionaryAspectsAnalyzed: dict[str, float | numpy.ndarray] = {}
	if 'aspectralstats' in FFprobeStructured:
		for keyName in FFprobeStructured['aspectralstats']:
			# No matter how many channels, each keyName is `numpy.ndarray[tuple[int, int], numpy.dtype[numpy.float64]]`
			# where `tuple[int, int]` is (channel, frame)
			# NOTE (as of this writing) `registrar` can only understand the generic class `numpy.ndarray` and not more specific typing
			dictionaryAspectsAnalyzed[keyName] = FFprobeStructured['aspectralstats'][keyName]
			# dictionaryAspectsAnalyzed[keyName] = numpy.mean(FFprobeStructured['aspectralstats'][keyName]).astype(float)
	if 'r128' in FFprobeStructured:
		for keyName in FFprobeStructured['r128']:
			dictionaryAspectsAnalyzed[keyName] = FFprobeStructured['r128'][keyName][-1]
	if 'astats' in FFprobeStructured:
		for keyName, arrayFeatureValues in cast(dict[str, numpy.ndarray[Any, Any]], FFprobeStructured['astats']).items():
			dictionaryAspectsAnalyzed[keyName.split('.')[-1]] = numpy.mean(arrayFeatureValues[..., -1:None]).astype(float)

	return dictionaryAspectsAnalyzed

@registrationAudioAspect('Zero-crossings rate')
def analyzeZero_crossings_rate(pathFilename: str | PathLike[Any]) -> float | None:
	return ffprobeShotgunAndCache(pathFilename).get('Zero_crossings_rate')

@registrationAudioAspect('DC offset')
def analyzeDCoffset(pathFilename: str | PathLike[Any]) -> float | None:
	return ffprobeShotgunAndCache(pathFilename).get('DC_offset')

@registrationAudioAspect('Dynamic range')
def analyzeDynamicRange(pathFilename: str | PathLike[Any]) -> float | None:
	return ffprobeShotgunAndCache(pathFilename).get('Dynamic_range')

@registrationAudioAspect('Signal entropy')
def analyzeSignalEntropy(pathFilename: str | PathLike[Any]) -> float | None:
	return ffprobeShotgunAndCache(pathFilename).get('Entropy')

@registrationAudioAspect('Duration-samples')
def analyzeNumber_of_samples(pathFilename: str | PathLike[Any]) -> float | None:
	return ffprobeShotgunAndCache(pathFilename).get('Number_of_samples')

@registrationAudioAspect('Peak dB')
def analyzePeak_level(pathFilename: str | PathLike[Any]) -> float | None:
	return ffprobeShotgunAndCache(pathFilename).get('Peak_level')

@registrationAudioAspect('RMS total')
def analyzeRMS_level(pathFilename: str | PathLike[Any]) -> float | None:
	return ffprobeShotgunAndCache(pathFilename).get('RMS_level')

@registrationAudioAspect('Crest factor')
def analyzeCrest_factor(pathFilename: str | PathLike[Any]) -> float | None:
	return ffprobeShotgunAndCache(pathFilename).get('Crest_factor')

@registrationAudioAspect('RMS peak')
def analyzeRMS_peak(pathFilename: str | PathLike[Any]) -> float | None:
	return ffprobeShotgunAndCache(pathFilename).get('RMS_peak')

@registrationAudioAspect('LUFS integrated')
def analyzeLUFSintegrated(pathFilename: str | PathLike[Any]) -> float | None:
	return ffprobeShotgunAndCache(pathFilename).get('I')

@registrationAudioAspect('LUFS loudness range')
def analyzeLRA(pathFilename: str | PathLike[Any]) -> float | None:
	return ffprobeShotgunAndCache(pathFilename).get('LRA')

@registrationAudioAspect('LUFS low')
def analyzeLUFSlow(pathFilename: str | PathLike[Any]) -> float | None:
	return ffprobeShotgunAndCache(pathFilename).get('LRA.low')

@registrationAudioAspect('LUFS high')
def analyzeLUFShigh(pathFilename: str | PathLike[Any]) -> float | None:
	return ffprobeShotgunAndCache(pathFilename).get('LRA.high')

@registrationAudioAspect('Power spectral density')
def analyzeMean(pathFilename: str | PathLike[Any]) -> numpy.ndarray:
	return ffprobeShotgunAndCache(pathFilename).get('mean')

@registrationAudioAspect('Spectral variance')
def analyzeVariance(pathFilename: str | PathLike[Any]) -> numpy.ndarray:
	return ffprobeShotgunAndCache(pathFilename).get('variance')

@registrationAudioAspect('Spectral centroid')
def analyzeCentroid(pathFilename: str | PathLike[Any]) -> numpy.ndarray:
	return ffprobeShotgunAndCache(pathFilename).get('centroid')

@registrationAudioAspect('Spectral spread')
def analyzeSpread(pathFilename: str | PathLike[Any]) -> numpy.ndarray:
	return ffprobeShotgunAndCache(pathFilename).get('spread')

@registrationAudioAspect('Spectral skewness')
def analyzeSkewness(pathFilename: str | PathLike[Any]) -> numpy.ndarray:
	return ffprobeShotgunAndCache(pathFilename).get('skewness')

@registrationAudioAspect('Spectral kurtosis')
def analyzeKurtosis(pathFilename: str | PathLike[Any]) -> numpy.ndarray:
	return ffprobeShotgunAndCache(pathFilename).get('kurtosis')

@registrationAudioAspect('Spectral entropy')
def analyzeSpectralEntropy(pathFilename: str | PathLike[Any]) -> numpy.ndarray:
	return ffprobeShotgunAndCache(pathFilename).get('entropy')

@registrationAudioAspect('Spectral flatness')
def analyzeFlatness(pathFilename: str | PathLike[Any]) -> numpy.ndarray:
	return ffprobeShotgunAndCache(pathFilename).get('flatness')

@registrationAudioAspect('Spectral crest')
def analyzeCrest(pathFilename: str | PathLike[Any]) -> numpy.ndarray:
	return ffprobeShotgunAndCache(pathFilename).get('crest')

@registrationAudioAspect('Spectral flux')
def analyzeFlux(pathFilename: str | PathLike[Any]) -> numpy.ndarray:
	return ffprobeShotgunAndCache(pathFilename).get('flux')

@registrationAudioAspect('Spectral slope')
def analyzeSlope(pathFilename: str | PathLike[Any]) -> numpy.ndarray:
	return ffprobeShotgunAndCache(pathFilename).get('slope')

@registrationAudioAspect('Spectral decrease')
def analyzeDecrease(pathFilename: str | PathLike[Any]) -> numpy.ndarray:
	return ffprobeShotgunAndCache(pathFilename).get('decrease')

@registrationAudioAspect('Spectral rolloff')
def analyzeRolloff(pathFilename: str | PathLike[Any]) -> numpy.ndarray:
	return ffprobeShotgunAndCache(pathFilename).get('rolloff')

@registrationAudioAspect('Abs_Peak_count')
def analyzeAbs_Peak_count(pathFilename: str | PathLike[Any]) -> float | None:
	print('Abs_Peak_count', pathFilename)
	return ffprobeShotgunAndCache(pathFilename).get('Abs_Peak_count')

@registrationAudioAspect('Bit_depth')
def analyzeBit_depth(pathFilename: str | PathLike[Any]) -> float | None:
	return ffprobeShotgunAndCache(pathFilename).get('Bit_depth')

@registrationAudioAspect('Flat_factor')
def analyzeFlat_factor(pathFilename: str | PathLike[Any]) -> float | None:
	return ffprobeShotgunAndCache(pathFilename).get('Flat_factor')

@registrationAudioAspect('Max_difference')
def analyzeMax_difference(pathFilename: str | PathLike[Any]) -> float | None:
	return ffprobeShotgunAndCache(pathFilename).get('Max_difference')

@registrationAudioAspect('Max_level')
def analyzeMax_level(pathFilename: str | PathLike[Any]) -> float | None:
	return ffprobeShotgunAndCache(pathFilename).get('Max_level')

@registrationAudioAspect('Mean_difference')
def analyzeMean_difference(pathFilename: str | PathLike[Any]) -> float | None:
	return ffprobeShotgunAndCache(pathFilename).get('Mean_difference')

@registrationAudioAspect('Min_difference')
def analyzeMin_difference(pathFilename: str | PathLike[Any]) -> float | None:
	return ffprobeShotgunAndCache(pathFilename).get('Min_difference')

@registrationAudioAspect('Min_level')
def analyzeMin_level(pathFilename: str | PathLike[Any]) -> float | None:
	return ffprobeShotgunAndCache(pathFilename).get('Min_level')

@registrationAudioAspect('Noise_floor')
def analyzeNoise_floor(pathFilename: str | PathLike[Any]) -> float | None:
	return ffprobeShotgunAndCache(pathFilename).get('Noise_floor')

@registrationAudioAspect('Noise_floor_count')
def analyzeNoise_floor_count(pathFilename: str | PathLike[Any]) -> float | None:
	return ffprobeShotgunAndCache(pathFilename).get('Noise_floor_count')

@registrationAudioAspect('Peak_count')
def analyzePeak_count(pathFilename: str | PathLike[Any]) -> float | None:
	return ffprobeShotgunAndCache(pathFilename).get('Peak_count')

@registrationAudioAspect('RMS_difference')
def analyzeRMS_difference(pathFilename: str | PathLike[Any]) -> float | None:
	return ffprobeShotgunAndCache(pathFilename).get('RMS_difference')

@registrationAudioAspect('RMS_trough')
def analyzeRMS_trough(pathFilename: str | PathLike[Any]) -> float | None:
	return ffprobeShotgunAndCache(pathFilename).get('RMS_trough')



from analyzeAudio import registrationAudioAspect, audioAspects, cacheAudioAnalyzers
from typing import Any
import cachetools
import librosa
import numpy
from optype.numpy import AnyFloatingDType, ToArray3D, ToFloat3D
from numpy import dtype, floating

@registrationAudioAspect('Chromagram')
def analyzeChromagram(spectrogramPower: numpy.ndarray[Any, dtype[floating[Any]]], sampleRate: int, **keywordArguments: Any) -> numpy.ndarray:
	return librosa.feature.chroma_stft(S=spectrogramPower, sr=sampleRate, **keywordArguments)

@registrationAudioAspect('Spectral Contrast')
def analyzeSpectralContrast(spectrogramMagnitude: numpy.ndarray[Any, dtype[floating[Any]]], **keywordArguments: Any) -> numpy.ndarray:
	return librosa.feature.spectral_contrast(S=spectrogramMagnitude, **keywordArguments)

@registrationAudioAspect('Spectral Bandwidth')
def analyzeSpectralBandwidth(spectrogramMagnitude: numpy.ndarray[Any, dtype[floating[Any]]], **keywordArguments: Any) -> numpy.ndarray:
	centroid = audioAspects['Spectral Centroid']['analyzer'](spectrogramMagnitude)
	return librosa.feature.spectral_bandwidth(S=spectrogramMagnitude, centroid=centroid, **keywordArguments)

@cachetools.cached(cache=cacheAudioAnalyzers)
@registrationAudioAspect('Spectral Centroid')
def analyzeSpectralCentroid(spectrogramMagnitude: numpy.ndarray[Any, dtype[floating[Any]]], **keywordArguments: Any) -> numpy.ndarray:
	return librosa.feature.spectral_centroid(S=spectrogramMagnitude, **keywordArguments)

@registrationAudioAspect('Spectral Flatness')
def analyzeSpectralFlatness(spectrogramMagnitude: numpy.ndarray[Any, dtype[floating[Any]]], **keywordArguments: Any) -> numpy.ndarray:
	spectralFlatness = librosa.feature.spectral_flatness(S=spectrogramMagnitude, **keywordArguments)
	return 20 * numpy.log10(spectralFlatness, where=(spectralFlatness != 0)) # dB
